/*

  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/

  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of authors nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*/

/* Locale dependent routines */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "ximserver.h"
#ifndef __GNUC__
# ifdef HAVE_ALLOCA_H
#  include <alloca.h>
# endif
#endif

static char *
ustring_to_utf8_str(uString *s)
{
    uString::iterator i;
    int l = 0, nbyte;
    char utf8[6];
    /* count the length */
    for (i = s->begin(); i != s->end(); i++) {
	nbyte = wctomb(utf8, *i);
	l += nbyte;
    }
    char *c = (char *)malloc(l + 1);
    c[l] = 0;
    l = 0;
    for (i = s->begin(); i != s->end(); i++) {
	nbyte = wctomb(utf8, *i);
	int j;
	for (j = 0; j < nbyte; j++) {
	    c[l] = utf8[j];
	    l++;
	}
    }
    return c;
}

bool
Locale::supportOverTheSpot()
{
    return false;
}

class EUCJP_Locale : public Locale {
public:
    virtual char *uStringToCtext(uString *us) {
	int len;
	char *t;
	len = us->size() * 10;
	t = (char *)alloca(len + 1);
	uString::iterator i;
	bool jp_mode = false;
	int cur, j = 0;

	for (i = us->begin(); i != us->end(); i++) {
	    cur = *i;
	    if (!(cur & 0xff00)) {
		if (jp_mode) {
		    t[j++] = 0x1b;
		    t[j++] = 0x28;
		    t[j++] = 0x42;
		    jp_mode = false;
		}
		t[j++] = cur & 0xff;
	    } else {
		if (!jp_mode) {
		    jp_mode = true;
		    t[j++] = 0x1b;
		    t[j++] = 0x24;
		    t[j++] = 0x29;
		    t[j++] = 0x42;
		}
		t[j++] = (cur >> 8) | 0x80;
		t[j++] = (cur & 0xff) | 0x80;
	    }
	}
	if (jp_mode) {
	    t[j++] = 0x1b;
	    t[j++] = 0x28;
	    t[j++] = 0x42;
	}
	t[j] = 0;
	return strdup(t);
    };
    virtual bool supportOverTheSpot() {
	return true;
    }
};

class UTF8_Locale : public Locale {
public:
    UTF8_Locale::UTF8_Locale(const char *lang);
    virtual char *uStringToCtext(uString *us) {
	char *str = ustring_to_utf8_str(us);
	XTextProperty prop;
	XmbTextListToTextProperty(XimServer::gDpy, &str, 1,
			      XCompoundTextStyle, &prop);
	free(str);
	char *res = strdup((char *)prop.value);
	XFree(prop.value);
	return res;
    }
    virtual bool supportOverTheSpot() {
	return true;
    }
    virtual void set_localename_from_lang(const char *lang);
    virtual const char *compose_localename_from_lang(const char *lang);
private:
    char *mLocaleName;
};

UTF8_Locale::UTF8_Locale(const char *lang)
{
    mLocaleName = strdup(compose_localename_from_lang(lang));
}

const char *
UTF8_Locale::compose_localename_from_lang(const char *lang)
{
    const char *name;
    // XXX Should we use tables? Also I haven't checked m17nlib yet.
    // --ekato
    if (!strcmp(lang, "ja"))
	name = "ja_JP.UTF-8";
    else if (!strcmp(lang, "ko"))
	name = "ko_KR.UTF-8";
    else if (!strcmp(lang, "zh_CN"))
	name = "zh_CN.UTF-8";
    else if (!strncmp(lang, "zh_TW",5)) // XXX use zh_TW.UTF-8 locale for Traditional Chinese IM
	name = "zh_TW.UTF-8";
    else if (!strcmp(lang, "vi"))
	name = "vi_VN.UTF-8";
    else
	name = "en_US.UTF-8";

    return name;
}

void
UTF8_Locale::set_localename_from_lang(const char *lang)
{
    const char *name;
    name = compose_localename_from_lang(lang);
    
    if (mLocaleName)
	free(mLocaleName);
    mLocaleName = strdup(name);
}

Locale *createLocale(const char *lang)
{
    return new UTF8_Locale(lang);
}
